package de.fraunhofer.iem.dataprovider.taskManager

import de.fraunhofer.iem.dataprovider.configuration.Config
import de.fraunhofer.iem.dataprovider.configuration.OpenCodeGitlabConfiguration
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.taskManager.events.*
import de.fraunhofer.iem.dataprovider.taskManager.tasks.dataQuery.CloneGitTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.dataQuery.GetGitlabProjectTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.dataQuery.GetRepositoryDetailsTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.kpiCalculation.MetricsTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.detekt.DetektTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.odc.OdcTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.qodana.QodanaTask
import de.fraunhofer.iem.dataprovider.toolRun.service.ToolRunService
import jakarta.annotation.PreDestroy
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ObsoleteCoroutinesApi
import kotlinx.coroutines.channels.SendChannel
import kotlinx.coroutines.channels.actor
import org.springframework.stereotype.Component
import java.nio.file.Paths
import java.util.*


/**
 * The Task Manager takes tasks and distributes them to
 * underlying workers. Internally it uses a channel
 * to manage incoming tasks.
 */
@Component
class TaskManager(
    private val config: Config,
    private val openCodeGitlabConfiguration: OpenCodeGitlabConfiguration,
    private val toolRunService: ToolRunService,
    private val repositoryService: RepositoryService,
    private val kpiService: KPIService
) {

    // The used default dispatcher is ok for CPU-bound workloads. However,
    // if they block for a long time it's better to use a custom thread
    // pool solution.
    private val worker = Worker("Worker")

    // Should be used for long-running tasks, which DON'T use CPU power and
    // are non-blocking
    private val ioWorker = Worker("IO-Worker", CoroutineScope(Dispatchers.IO))
    // Should be used for long-running CPU heavy tasks, which potentially block.
    // private val customWorker = getCustomThreadpoolWorker()

    private val mainScope = CoroutineScope(Dispatchers.Default)
    private val actor: SendChannel<Event> = mainScope.taskActor()

    suspend fun addEvent(event: Event) {
        actor.send(event)
    }

    @OptIn(ObsoleteCoroutinesApi::class)  // This API is meant to be improved since 2017... so this should be good to use.
    // IntelliJ might tell u that it is ok to remove <ITask> from the actor declaration.
    // IT IS NOT! this will lead to very strange gradle compile errors!
    private fun CoroutineScope.taskActor() = actor<Event> {
        with(MyActor()) {
            for (task in channel) {
                onReceive(task)
            }
        }
    }

    @PreDestroy
    fun close() {
        actor.close()
        worker.close()
        ioWorker.close()
    }

    /**
     * The state in this internal class is thread safe.
     * See https://kotlinlang.org/docs/shared-mutable-state-and-concurrency.html#actors
     * It is used to manage e.g. fixed task sequences, as well as react to general events.
     */
    private inner class MyActor {

        private val logger = getLogger(javaClass)
        private val dependentEvents: MutableMap<UUID, MutableSet<UUID>> = mutableMapOf()
        suspend fun onReceive(event: Event) {
            logger.info("[${Thread.currentThread().name}] add task called in Task Manager $event")

            when (event) {
                is RepoChangedEvent -> {
                    ioWorker.addTask(
                        GetGitlabProjectTask(
                            event.repoId,
                            openCodeGitlabConfiguration,
                            ::addEvent,
                            repositoryService,
                        )
                    )
                }

                is GetGitlabProjectDoneEvent -> {

                    ioWorker.addTask(
                        CloneGitTask(
                            event.gitRepository,
                            event.repoId,
                            ::addEvent,
                            config.gitProjectPath
                        )
                    )
                }

                is GitCloneDoneEvent -> {
                    val taskIds = mutableSetOf<UUID>()
                    val groupId = UUID.randomUUID()
                    // TODO: analysis results path should be unique or best not placed in the git folder
                    val analysisResultsPath = Paths.get(event.outputDirectory, "tool-results").toString()

                    val odcTask = OdcTask(
                        event.outputDirectory,
                        analysisResultsPath,
                        ::addEvent,
                        event.repoId,
                        groupId,
                        toolRunService
                    )

                    val detektTask = DetektTask(
                        event.outputDirectory,
                        analysisResultsPath,
                        ::addEvent,
                        event.repoId,
                        groupID = groupId,
                        toolRunService
                    )

                    val qodanaTask = QodanaTask(
                        event.outputDirectory,
                        analysisResultsPath,
                        ::addEvent,
                        event.repoId,
                        groupID = groupId,
                        toolRunService
                    )

                    taskIds.add(odcTask.taskID)
                    taskIds.add(detektTask.taskID)
                    taskIds.add(qodanaTask.taskID)

                    dependentEvents[groupId] = taskIds

                    worker.addTask(odcTask)
                    worker.addTask(detektTask)
                    worker.addTask(qodanaTask)
                }

                is SarifProcessGroupDoneEvent -> {
                    logger.info("Done with Sarif Task")
                    val eventGroup = dependentEvents[event.groupId]
                    if (eventGroup != null) {
                        eventGroup.remove(event.taskId)
                        if (eventGroup.isEmpty()) {
                            logger.info("Adding repository details task")
                            worker.addTask(
                                GetRepositoryDetailsTask(
                                    event.repoId,
                                    openCodeGitlabConfiguration,
                                    repositoryService,
                                    ::addEvent
                                )
                            )
                        }
                    }
                }

                is GetRepositoryDetailsDoneEvent -> {
                    worker.addTask(MetricsTask(event.repoId, repositoryService, kpiService, ::addEvent))
                }

                else -> {
                    logger.info("Received event without special handling associated $event")
                }
            }
            logger.info("[${Thread.currentThread().name}] add task finished in Task Manager $event")
        }

    }
}
