package de.fraunhofer.iem.dataprovider.taskManager.tasks.kpiCalculation

import de.fraunhofer.iem.dataprovider.kpi.dto.KPIDto
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.kpi.strategy.AggregationKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RatioKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RawValueKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.repository.entity.RepositoryDetailsEntity
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.taskManager.events.Event
import de.fraunhofer.iem.dataprovider.taskManager.tasks.Task
import java.util.*

class MetricsTask(
    val repoId: UUID,
    private val repositoryService: RepositoryService,
    private val kpiService: KPIService,
    override val responseChannel: suspend (task: Event) -> Unit
) : Task() {
    override suspend fun execute() {
        logger.info("Starting metrics task for repoId $repoId")
        val repository = repositoryService.findRepoByID(repoId)
        val repositoryDetails = repositoryService.getLatestRepositoryDetailsByRepositoryId(repoId)
        if (repositoryDetails != null && repository != null) {
            val rootKPI = generateKPITree(repositoryDetails)
            kpiService.calculateKPIs(rootKPI)
            kpiService.storeAndPurgeOld(repository, rootKPI)
        }
    }

    private fun generateKPITree(repositoryDetailsEntity: RepositoryDetailsEntity): KPIDto {
        // lowest level leaves
        val numberOfCommitsKPI = KPIDto(
            "Number of commits",
            "Overall transparency score between 0-100 :-)",
            false,
            RawValueKPICalculationStrategy(repositoryDetailsEntity.numberOfCommits!!)
        )
        val numberOfSignedCommitsKPI = KPIDto(
            "Number of signed commits",
            "Overall transparency score between 0-100 :-)",
            false,
            RawValueKPICalculationStrategy(repositoryDetailsEntity.numberOfSignedCommits!!)
        )
        val isDefaultBranchProtectedKPI = KPIDto(
            "Is Default Branch Protected",
            "Is the default branch protected?",
            false,
            RawValueKPICalculationStrategy(if (repositoryDetailsEntity.isDefaultBranchProtected == true) 100 else 0)
        )

        val signedCommitsRatioKPI =
            KPIDto("Signed Commit Ratio", "Ratio between signed and all commits", false, RatioKPICalculationStrategy())
        signedCommitsRatioKPI.addChildKPI(numberOfCommitsKPI, 1.0)
        signedCommitsRatioKPI.addChildKPI(numberOfSignedCommitsKPI, 1.0)

        // second level
        val securityKPI = KPIDto(
            "Security score",
            "Overall project security score between 0-100 :-)",
            false,
            AggregationKPICalculationStrategy()
        )
        val transparencyKPI = KPIDto(
            "Transparency score",
            "Overall transparency score between 0-100 :-)",
            false,
            AggregationKPICalculationStrategy()
        )
        securityKPI.addChildKPI(isDefaultBranchProtectedKPI, 0.5)
        securityKPI.addChildKPI(signedCommitsRatioKPI, 0.5)
        transparencyKPI.addChildKPI(signedCommitsRatioKPI, 1.0)

        val rootKPI = KPIDto( "Project score", "Overall project score between 0-100 :-)", true, AggregationKPICalculationStrategy())
        rootKPI.addChildKPI(securityKPI, 0.5)
        rootKPI.addChildKPI(transparencyKPI, 0.5)
        return rootKPI
    }
}
